require 'redis'
require 'json'
require_relative 'service_controller'

module Praetor
  class Service
    class << self
      attr_reader :_controllers

      def controller(klass)
        klass.include ServiceController
        (@_controllers ||= []) << klass
      end

      def endpoint
        @endpoint ||= name.underscore
      end
    end

    attr_accessor :redis

    def run
      client_options = ENV['REDIS_PROVIDER'] ? { url: ENV['REDIS_PROVIDER'] } : {}
      @redis = Redis.new client_options

      loop do
        # Pop off the next request (or block until one arrives)
        _, request = @redis.brpop self.class.endpoint

        # You can shut down a service by issuing a stop command; useful for testing
        break if request == 'stop'
        handle_request JSON.parse(request)
      end
    end

    def dispatch!(request)
      controller = request['controller']
      action     = request['action']
      params     = request['params'] || {}

      check_controller! controller

      klass = controller.constantize
      check_action! klass, action

      instance = klass.new
      instance.instance_variable_set :@params, params
      instance.send action
    end

    private

    def check_controller!(controller)
      unless self.class._controllers.map(&:name).include? controller
        raise "No controller that matches `#{controller}`"
      end
    end

    def check_action!(controller, action)
      unless controller.method_defined? action
        raise "No action `#{action}` for controller `#{controller}`"
      end
    end

    def handle_request(request)
      response = begin
        data = dispatch! request
        { 'status' => 'ok', 'data' => data }
      rescue Exception => e
        {
          'status' => 'error',
          'data'   => { 'message' => e.message, 'backtrace' => e.backtrace }
        }
      end

      respond response.merge('key' => request['key'])
    end

    def respond(response)
      # Push the response onto a new list using the request/response key
      @redis.rpush response['key'], JSON.generate(response)

      # Set an expire time for our list to make sure we don't leak
      @redis.expire response['key'], 30
    end
  end
end
